# nopickles.py is copyright 2014 Spiky Caterpillar and released under the
# WTFPL.

# These are the rather hackish routines to replace Ren'Py's use of Python
# pickles.  Note that while this is a lot more secure than pickle, it's not
# currently fully secure.  In particular, note that it ships with
# INSECURE=True set.

# Before you panic, note that the only way I know of to actually exploit
# the pickles is for someone to have write access to your saved games or
# persistent data - so things like trading saves and persistent, or letting
# someone/something else write your files, or using networked providers
# to store your saves.  So it's a real risk, but only a HUGE one if I were
# using some form of networked save system.

# Actually using this requires some minor changes to Ren'Py (diffing the
# Date Warp Ren'Py against a stock 6.17.7 should tell you what these changes
# are) and constraining your code to only the syntax that the nopickles
# inspector handles.  If you're seriously planning on using it, you may want
# to talk it over with Spiky for integration advice.

import renpy
import unroller
import os
from namedtransform import *
import sanitize
import sanitize_bc
#import animation

# Used in nopickles.rpy to determine whether or not to dump jpegs.
SAVE_SCREENSHOTS = False

INSECURE = False
DEBUG = True

# If set, autosave on every single line for debugging purposes.
AUTOHAMMER = False

FUTUREDEBUG = True # Set to True to provide debug information about things
# which are not a problem for Date Warp but may be a problem for future games.

# Warnings are a list of sequences; the first item of the sequence is the
# severity, the second and subsequent items are data.
#
# Severities:
# 5: Discarded data.
# 6: Accepted tainted data.
class Warnings(list):
    def __init__(self):
        list.__init__(self)
        self.notes = []
    def note(self,v):
        self.notes.append(v)
    def unnote(self):
        self.notes.pop()
    def append(self, v):
        list.append(self,v+tuple(self.notes))

def clear_warnings():
    global warnings
    warnings = Warnings()

clear_warnings()

def most_important_warning():
    highest = -1
    for warning in warnings:
        if warning[0]>highest:
            highest = warning[0]
            highest_val = warning
    if highest>=0:
        ret = repr(highest_val)
        return ret.replace('{','{{')
    return '(No warnings.)'

def serialize_dict_contents(item,rollobjmemo=None):
    ret = []
    for k, v in item.iteritems():
        ret.extend(serialize_value(k,rollobjmemo=rollobjmemo))
        ret.extend(serialize_value(v,rollobjmemo=rollobjmemo))
    return ret

def serialize_value(item,memoset=None,rollobjmemo=None):
    ti = type(item)
    if ti is tuple:
        ret = ['(',]
        for i in item:
            ret.extend(serialize_value(i,rollobjmemo=rollobjmemo))
        ret.append(')')
        return ret
    elif item is None:
        return 'None',
    elif ti is str or ti is unicode:
        return 'str '+escape_str(item),
    elif ti is int or ti is long:
        return 'int '+str(item),
    elif ti is float:
        return 'float '+str(item),
    elif ti is bool:
        return str(item),
    elif ti is dict:
        ret = ['p{',]+serialize_dict_contents(item,rollobjmemo=rollobjmemo)+['}',]
        return ret
    elif ti is set:
        ret = ['s(']
        for member in item:
            ret.extend(serialize_value(member))
        ret.append(')')
        return ret
    elif ti is renpy.python.RevertableDict:
        ret = ['{']
        ret.extend(serialize_dict_contents(item,rollobjmemo=rollobjmemo))
        return ret + ['}',]
    elif ti is renpy.python.RevertableList:
        ret = ['[',]
        for i in item:
            ret.extend(serialize_value(i,memoset=memoset,rollobjmemo=rollobjmemo))
        return ret + [']',]
    elif ti is NamedTransform:
       return ('trans '+item.name,)
    #elif animation.AnimController in ti.mro(): DateWarp-specific, so removed.
    #    return ('ac '+ti.__name__,)
    elif isinstance(item,SanitizableRevertableObject):
        return item.Serialize(memoset,rollobjmemo=rollobjmemo)
    print "BUG: Cannot serialize values of type",type(item)
    if AUTOHAMMER:
        print item
        raise Exception('Serialization error.')
    return ('#.44 unhandled value'+str(type(item)),)

def serialize_tnv(k, v, memoset=None, memoid=None, rollobjmemo=None):
    # Serialize a key/value pair to a sequence of one or more strings.
    # The format is type, name, value.
    # Non-scalar values will extend over multiple strings in the returned
    # sequence.
    if rollobjmemo is None:
        print "Warning: Serializer",k,v,memoid
    if type(v) is str or type(v) is unicode:
        return 'str '+k+' '+escape_str(v),
    elif type(v) is int:
        return 'int '+k+' '+str(v),
    elif type(v) is bool:
        return str(v)+' '+k,
    elif type(v) is float:
        return 'float '+k+' '+str(v),
    elif type(v) is tuple:
        ret = ['( '+k]
        for i in v:
            ret.extend(serialize_value(i, memoset=memoset,rollobjmemo=rollobjmemo))
        ret.append(')')
        return ret
    elif v is None:
        return 'None '+k,
    elif type(v) is renpy.python.RevertableDict:
        ret = ['{ '+k]
        for dk, dv in v.iteritems():
            ret.extend(serialize_value(dk,rollobjmemo=rollobjmemo))
            ret.extend(serialize_value(dv,memoset=memoset,rollobjmemo=rollobjmemo))
        ret.append('}')
        return ret
    elif type(v) is dict:
        ret = ['p{ '+k]
        for dk, dv in v.iteritems():
            ret.extend(serialize_value(dk,rollobjmemo=rollobjmemo))
            ret.extend(serialize_value(dv,memoset=memoset,rollobjmemo=rollobjmemo))
        ret.append('}')
        return ret
    elif type(v) is renpy.python.RevertableList:
        ret = ['[ '+k]
        if k=='pending_actors':
            if rollobjmemo is not None:
                if id(v) in rollobjmemo:
                    ret.append('#.167 '+str(rollobjmemo[id(v)]))
                else:
                    #for subk, subv in rollobjmemo.iteritems():
                    #    ret.append('#.171'+str(subk)+';'+str(subv))
                    ret.append('#:171 '+str(id(v))+';'+str(memoid))
            ret.append('#.169'+str(v))
        if rollobjmemo is None:
            ret.append('#.160: obmemo is None')
        else:
            rolledv = renpy.python.RevertableList()
            if id(v) in rollobjmemo:
                rolledv.rollback(rollobjmemo[id(v)])
                ret.append("#.179"+str(rolledv))
            else:
                ret.append("#.166")
            ret.append("#.164 "+str(rolledv))
        for i in v:
            ret.extend(serialize_value(i,memoset=memoset,rollobjmemo=rollobjmemo))
        ret.append(']')
        return ret
    elif type(v) is renpy.python.RevertableSet:
        ret = ['rs( '+k]
        for i in v:
            ret.extend(serialize_value(i,rollobjmemo=rollobjmemo))
        ret.append(')')
        return ret
    elif type(v) is set:
        ret = ['s( '+k]
        for i in v:
            ret.extend(serialize_value(i,rollobjmemo=rollobjmemo))
        ret.append(')')
        return ret
    elif type(v) is renpy.python.StoreDeleted:
        return ('del '+k,)
    elif isinstance(v, SanitizableRevertableObject):
        return v.Serialize(memoset,infix=' '+k,memoid=memoid, rollobjmemo=rollobjmemo)
    else:
        return '# Invalid tnv: '+k+':'+str(type(v)),
        
def serialize_script_loc(loc):
    # We reverse the order of locations so they're line, timestamp, file
    # so that files with spaces will not screw up parsing.
    if type(loc) is not tuple:
        if type(loc) is str:
            return 'lbl '+escape_str(loc)
        serialize_warnings.append("Not sure how to serialize location "+repr(loc))
    if len(loc)!=3:
        serialize_warnings.append("Script location "+repr(loc)+" length unexpected.")
    return str(loc[2])+' '+str(loc[1])+' '+str(loc[0])

def serialize_rollback_log(log):
    roots = renpy.game.log.freeze()
    #roots = log.get_roots()
    ret = ['log',]
    found = False
    l = list(log.log)
    l.reverse()
    elongatable = []
    for ent in l:
        #ret.append("#.156: "+str(roots))
        es = SerializedRollbackLogEntry(ent, roots)
        if es.can_be_last():
            roots = False
            found = True
        elongatable.append(es)
    elongatable.reverse()
    for es in elongatable:
        ret.extend(es.elongate())
    ret.append('end')
    if not found:
        raise Exception("No valid serialization points found.")
    return ret

def serialize_context_musicchannel(channelname, data):
    ret = []
    if data.last_filenames:
        for fn in data.last_filenames:
            ret.append('file '+escape_str(fn))
        if data.secondary_volume!=1.0:
            ret.append('vol '+str(data.secondary_volume))
        if data.pan!=0:
            print "Warning: nopickles does not currently save pan values."
    if ret:
        return ['channel '+str(channelname),]+ret+['end',]
    return ()

def serialize_context_music(music):
    # A context's music is a Ren'Py RevertableDict.  I'm not entirely sure why.
    ret = []
    for k,v in music.iteritems():
        ret.extend(serialize_context_musicchannel(k,v))
    if ret:
        return ['music',]+ret+['end',]
    return ()

ROLLBACK_OBJECT_TYPECODES={
    renpy.python.RevertableObject:'ro'
    }

def serialize_rollback_object(obj, data):
    to = type(obj)
    if to not in ROLLBACK_OBJECT_TYPECODES:
        return '#.155'+str(type(obj)),
    ret = ["%x"%id(obj)+" "+ROLLBACK_OBJECT_TYPECODES[to],]
    return ret+serialize_value(data)


class SerializedRollbackLogEntry(object):
    location_makes_sense = False
    has_roots = False
    def __init__(self, entry, roots):
        self.renpy_entry = entry
        cloc = entry.context.current
        # Only return members of the rollback log that make sense - experimental
        if type(cloc) is not tuple:
            return
        if len(cloc) != 3:
            return
        if type(cloc[2]) is not int or type(cloc[1]) is not int or type(cloc[0]) is not unicode:
            return
        self.location_makes_sense = True
        self.data = ret = ['entry','callstack']
        memoset = set()
        i = 0
        if DEBUG:
            if len(entry.context.call_location_stack)!=len(entry.context.return_stack):
                raise Exception("Error: call stack size mismatch.")
        while i<len(entry.context.call_location_stack):
            ret.append(serialize_script_loc(entry.context.call_location_stack[i]))
            rse = entry.context.return_stack[i]
            if type(rse) is tuple:
                ret.append('ret '+serialize_script_loc(rse))
            else:
                ret.append('retlbl '+escape_str(rse))
            i += 1
        ret.append('end')
        ret.extend(serialize_context_music(entry.context.music))
        objmemo = {}
        for k,v in entry.objects:
            objmemo[id(k)] = v
        for item in entry.context.dynamic_stack:
            # These are regular dicts.
            ret.append('dynamic')
            for k,v in item.iteritems():
                if k not in SERIALIZE_STORE_OVERRIDES:
                    ret.extend(serialize_tnv(k,v,memoset=memoset,rollobjmemo=objmemo))
                else:
                    ret.append('# Debug: 303')
                    ret.extend(SERIALIZE_STORE_OVERRIDES[k])
            ret.append('end')
        ret.append('current '+serialize_script_loc(cloc))
        ret.append('#168.')
        for storename, storeval in entry.stores.iteritems():
            if storeval:
                ret.append('store '+storename)
                for k,v in storeval.iteritems():
                    if k not in SERIALIZE_STORE_OVERRIDES:
                        ret.extend(serialize_tnv(k,v,memoset=memoset,rollobjmemo=objmemo))
                    else:
                        ret.append('#.315')
                        ret.extend(SERIALIZE_STORE_OVERRIDES[k])
                ret.append('end')
        # Behind experimental rollback-related saving.
        #if entry.objects:
        #    ret.append('objects')
        #    for obj, data in entry.objects:
        #       ret.extend(serialize_rollback_object(obj, data))
        #    ret.append('end')
        # End experimental rollback-related saving.
        self.root_chunk = ['#.root chunk']
        try:
            if roots:
                for k,v in roots.iteritems():
                    #try:
                    #    print "Debug:",k,v
                    #except Exception,e:
                    #    print "Exception printing debug message:",e
                    if id(v) in objmemo:
                        self.root_chunk.append('#.211 rollback '+str(v)+'->'+str(objmemo[id(v)]))
                        if isinstance(v, SanitizableRevertableObject):
                            nv = v.RolledBackClone(objmemo[id(v)])
                        else:
                            try:
                                nv = type(v)()
                                nv.rollback(objmemo[id(v)])
                                self.root_chunk.append('#:332'+str(nv))
                                self.root_chunk.append('#:333'+str(v))
                            except Exception,e:
                                nv = None
                                raise Exception("An exception occurred serializing "+repr(k)+' '+str(type(v))+" "+repr(v)+": "+str(e))
                        self.root_chunk.extend(serialize_tnv(k,nv,memoset=memoset,memoid=id(v),rollobjmemo=objmemo))
                    elif k in SERIALIZE_STORE_OVERRIDES:
                        self.root_chunk.extend(SERIALIZE_STORE_OVERRIDES[k])
                    else:
                        self.root_chunk.extend(serialize_tnv(k,v,memoset=memoset,rollobjmemo=objmemo))
            self.has_roots = True
            self.root_chunk.append('#/root chunk')
        except Exception,e:
            if DEBUG:
                print "Debug: Could not serialize properly.",e
                raise
            self.root_exception = e
        ret.extend(serialize_scene_lists(entry.context.scene_lists))
    def can_be_last(self):
        return self.location_makes_sense and self.has_roots

    # removed objects_chunk because it's rendered obsolete by the fact that
    # we actually roll back things in the saves.

    def elongate(self):
        if not self.location_makes_sense:
            return ()
        if self.has_roots:
            return self.data + self.root_chunk + ['end',]
        return self.data+['end',]

alphalow = 'abcdefghijklmnopqrstuvwxyz'

def clean_string(s, allow, replace='?'):
    ret = ''
    for c in s:
        if c in allow:
            ret += c
        else:
            ret += replace
    return ret

def unserialize_value(context_stack, line):
    if line=='(':
        context_stack.append(SafeloadTupleValueContext())
        return context_stack[-1]
    elif line=='[':
        context_stack.append(SafeloadRevertableListValueContext())
        return context_stack[-1]
    elif line=='p{':
        context_stack.append(SafeloadPythonDictValueContext())
        return context_stack[-1]
    elif line=='{':
        context_stack.append(SafeloadRevertableDictValueContext())
        return context_stack[-1]
    elif line=='s(':
        context_stack.append(SafeloadPythonSetValueContext())
        return context_stack[-1]
    elif line=='False':
        return False
    elif line=='None':
        return None
    elif line=='True':
        return True
    elif line.startswith('float '):
        return float(line[6:])
    elif line.startswith('int '):
        return int(line[4:])
    elif line.startswith('sro '):
        context_stack.append(SafeloadSanitizableRevertableObjectContext(line[4:]))
        return context_stack[-1]
    elif line.startswith('sroref '):
        memoctx = get_stack_memo_context(context_stack)
        return memoctx.OBJECT_MEMO[i64dec(line[7:])]
    elif line.startswith('str '):
        return unescape_str(line[4:])
    elif line.startswith('trans '):
        trans = getattr(renpy.store,line[6:],None)
        if type(trans) is NamedTransform:
            return trans
        else:
            warnings.append((5,"unserialize_value discarded missing transform",trans))
            ret = getattr(renpy.store,'unresolvable')
            if type(ret) is not NamedTransform:
                raise Exception("renpy.store.default is not a NamedTransform.")
            return ret
    elif line.startswith('ac '):
        candidate = getattr(animation,line[3:])
        if type(candidate) is type(animation.AnimController) and animation.AnimController in candidate.mro():
            return candidate()
        warnings.append((5,"unserialize_value used NullAnimController for",line))
        return animation.NullAnimController()
        
    raise Exception("Debug: unserialize_value discarded '"+line+"'")
    return None

class DeindirectableContext(object):
    """
    Used to check to see whether something should be deindirected before use.

    Note that deindirection is meant to be recursive; a deindirected value should
    contain no deindirectable values.
    """
    pass

class SafeloadSequenceValueContext(DeindirectableContext):
    """
    This isn't meant to be used directly; rather, it's a superclass for
    tuples and lists.
    """
    def __init__(self):
        self.l = []

    def deindirect_values(self):
        i = 0
        while i<len(self.l):
            if isinstance(self.l[i], SafeloadSequenceValueContext):
                self.l[i] = self.l[i].deindirect()
            if type(self.l[i]) is SafeloadPythonDictValueContext:
                self.l[i] = self.l[i].deindirect()
            i += 1

class SafeloadRevertableListValueContext(SafeloadSequenceValueContext):
    def handle(self,context_stack,line):
        if line==']':
            context_stack.pop()
            self.deindirect_values()
            return
        self.l.append(unserialize_value(context_stack, line))
    def deindirect(self):
        return renpy.python.RevertableList([deindirect_if_needed(i) for i in self.l])

class SafeloadPythonDictValueContext(DeindirectableContext):
    def __init__(self):
        self.l = []
    def handle(self, context_stack, line):
        if line=='}':
            context_stack.pop()
            return
        self.l.append(unserialize_value(context_stack, line))
    def build_dict(self):
        i = 0
        d = {}
        while i<len(self.l):
            d[deindirect_if_needed(self.l[i])] = deindirect_if_needed(self.l[i+1])
            i += 2
        return d
    def deindirect(self):
        return self.build_dict()

class SafeloadRevertableDictValueContext(SafeloadPythonDictValueContext):
    def build_dict(self):
        raise Exception("Wrong type.")
    def build_revertable_dict(self):
        i = 0
        d = renpy.python.RevertableDict()
        while i<len(self.l):
            d[deindirect_if_needed(self.l[i])] = deindirect_if_needed(self.l[i+1])
            i += 2
        return d
    def deindirect(self):
        return self.build_revertable_dict()

def get_stack_memo_context(stack):
    ret = None
    for ent in stack:
        if type(ent) is SafeloadRollbackLogEntryContext:
            if ret is None:
                ret = ent
            else:
                warnings.append((4,'RollbackLogEntry inside RollbackLogEntry, using outermost for memo context.'))
    return ret

class SafeloadSanitizableRevertableObjectContext(DeindirectableContext):
    actual_object = None
    sanitizable = True
    def __init__(self, typename):
        for c in typename:
            if c not in OK_VARNAME_FIRST:
                raise Exception("Disallowed sanitizable revertable object context type name.")
        t = getattr(renpy.store,typename)
        if (type(t) is not type(SanitizableRevertableObject)) or (SanitizableRevertableObject not in t.mro()):
            self.sanitizable = False
            warnings.append((5,typename,'is not a SanitizableRevertableObject type'))
        self.typename = typename
        self.l = []
    def handle(self, context_stack, line):
        if line=='end':
            context_stack.pop()
            return
        elif line.startswith('id '):
            if not self.sanitizable:
                warnings.append((5,"Discarded non-sanitizable object id",line))
                return
            memoctx = get_stack_memo_context(context_stack)
            self.objid = i64dec(line[3:])
            if self.objid in memoctx.OBJECT_MEMO:
                warnings.append((5,'ERROR: cloned object',self.typename,self.objid,tuple(context_stack)))
            memoctx.OBJECT_MEMO[self.objid] = self
            return
        self.l.append(parse_tainted_tnv(line, context_stack))
    def deindirect(self):
        if self.actual_object is not None:
            return self.actual_object
        elif not self.sanitizable:
            self.actual_object = renpy.python.deleted
            warnings.append((5,'Deindirection of bogus type',self.typename,'returning deleted'))
            return renpy.python.deleted
        ret_type = getattr(renpy.store, self.typename)
        ret = ret_type.__new__(ret_type)
        self.actual_object = ret
        warnings.note(ret_type.__name__)
        for item in self.l:
            # This will crash if the sanitizer does not exist, so is safe.
            warnings.note(item.name)
            try:
                setattr(ret, item.name, ret._member_sanitizers_[item.name](deindirect_if_needed(item.val)))
            except:
                print "Error running",item.name,ret._member_sanitizers_[item.name],item.val,deindirect_if_needed(item.val)
                raise
            warnings.unnote()
        ret.AfterLoad()
        warnings.unnote()
        return ret

DIRECT_TYPES = (bool, float, int, long, type(None), str, unicode, tuple,
    renpy.python.RevertableList,)

def deindirect_if_needed(v):
    if isinstance(v, DeindirectableContext):
        return v.deindirect()
    if isinstance(v,SanitizableRevertableObject):
        return v
    if type(v) not in DIRECT_TYPES:
        raise Exception("Unexpected type "+str(type(v)))
    return v

class SafeloadTupleValueContext(SafeloadSequenceValueContext):
    def __init__(self):
        self.l = []
    def handle(self, context_stack, line):
        if line == ')':
            context_stack.pop()
            return
        self.l.append(unserialize_value(context_stack, line))
    def deindirect(self):
        ret = tuple([deindirect_if_needed(i) for i in self.l])
        return ret

class SafeloadRevertableSetValueContext(SafeloadSequenceValueContext):
    def handle(self, context_stack, line):
        if line==')':
            context_stack.pop()
            return
        self.l.append(unserialize_value(context_stack, line))
    def deindirect(self):
        return renpy.python.RevertableSet([deindirect_if_needed(i) for i in self.l])

class SafeloadPythonSetValueContext(SafeloadSequenceValueContext):
    def handle(self, context_stack, line):
        if line==')':
            context_stack.pop()
            return
        self.l.append(unserialize_value(context_stack, line))
    def deindirect(self):
        return set([deindirect_if_needed(i) for i in self.l])

PERMITTED_SCRIPT_LOCATION_DIRS = (
    # These are directories where script locations can be loaded from (e.g.
    # call stack, current location, return destinations.)
    '/home/renpy/BlackCloset/game',
    '/home/renpy/projects/BlackCloset/game',
    '/home/renpy/projects/BlackClosetDemo/game',
    )

def unserialize_script_loc(line):
    loc = line.split(' ',2)
    if loc[0]=='lbl':
        if len(loc)!=2:
            warnings.append((5, 'Discarded unhandled script location',loc))
            return 'nopickles_replacement_label'
        loc = unescape_str(loc[1])
        if loc in ALLOWED_LOCATION_LABELS:
            return loc
        warnings.append((5, 'Discarded unhandled script location',loc))
        return 'nopickles_replacement_label'
    # Provisional sanity checking of locations in as of 2015 Apr 16.
    locsplit = loc[2].rsplit('/',1)
    if locsplit[0] not in PERMITTED_SCRIPT_LOCATION_DIRS:
        if locsplit[0] == '/home/tom/ab/renpy/renpy/common':
            # This stops saves from putting locations from the standard Ren'Py
            # library onto the call stack.  This is good because it saves me
            # from having to security-audit all the rpys in common/.
            #
            # nopickles_replacement_label will give a user-visible warning
            # if the game ever actually jumps there; we don't add a note
            # to the on-load warnings[] list because Ren'Py places stdlib
            # locations in unreachable-but-saved places on the call stack
            # sometimes.
            if AUTOHAMMER:
                # Wrapped in autohammer because a normal save will have
                # a bunch of these.
                print "Debug: unserialize_script_loc got disallowed",repr(loc)
            return 'nopickles_replacement_label'
        warnings.append((5,"Discarded unexpected script location", loc, locsplit))
        return 'nopickles_replacement_label'
    return (loc[2], int(loc[1]), int(loc[0]))

ALLOWED_LOCATION_LABELS = set((
    'nopickles_replacement_label',
    ))

class SafeloadCallstackContext(object):
    def __init__(self):
        self.call_location_stack = []
        self.return_stack = []
    def handle(self, context_stack, line):
        if line=='end':
            context_stack.pop()
            return
        elif line.startswith('ret '):
            self.return_stack.append(unserialize_script_loc(line[4:]))
        elif line.startswith('retlbl '):
            if INSECURE:
                # XXX need to scrub retlbls.
                warnings.append((6,'accepted',line))
                self.return_stack.append(unescape_str(line[7:]))
            elif line[7:] in ALLOWED_LOCATION_LABELS:
                self.return_stack.append(line[7:])
            else:
                warnings.append((5,'discarded',line))
                self.return_stack.append('nopickles_stack_error')
        else:
            self.call_location_stack.append(unserialize_script_loc(line))

def scrub_g_points(s):
    ret = renpy.python.RevertableDict()
    if type(s) is renpy.python.StoreDeleted:
        return renpy.python.deleted
    if type(s) is not renpy.python.RevertableDict:
        warnings.append((5,"Expected a RevertableDict, got",type(s)))
        return ret
    for k,v in s.iteritems():
        if type(k) is str and type(v) is int:
            ret[k] = v
    return ret

SCREEN_SCOPE_SCRUBBERS = dict(
    bigcalendar = dict(
        hint = sanitize.expect_displayable_text
        ),
    checklist_notify = dict(
        message = sanitize.expect_displayable_text
        )
    )

SERIALIZE_STORE_OVERRIDES = {
    # If present, the following constants will replace serialize_tnv for
    # the store (and only the store.)
    'store.current_bigcard' : ('None store.current_bigcard',),
    'store.forced_cases' : ('[ store.forced_cases', ']'),
    'store.moving_cards' : ('[ store.moving_cards',']'),
    'moving_cards' : ('[ moving_cards', ']'),
    'store.updater_target_version' : ('# store.updater_target_version',),
    }

STORE_SCRUBBERS = dict(
    # Ren'Py internals and non-bc-specific Spiky features are added later,
    # to work around the 255-argument limit.
    _a_trace_of_scarlet = sanitize_bc.scrub__a_trace_of_scarlet,
    a_pot_of_honey = sanitize_bc.expect_Case_instance_or_del,
    actor = sanitize_bc.expect_Student_instance_or_None, # For plotless.
    all_students_ever = sanitize_bc.scrub_all_students_ever,
    allow_save = sanitize.expect_bool,
    althea_ = sanitize_bc.expect_Councillor_or_del,
    althea_april_disappear_day = sanitize.expect_del_or_int,
    althea_caught_up_with_fangirl = sanitize_bc.scrub_althea_caught_up_with_fangirl,
    althea_coffeehouse = sanitize.expect_bool,
    althea_confession_loyalty = sanitize.expect_bool_or_None,
    althea_date_response = sanitize.expect_displayable_text_or_None,
    althea_embarrassing_2 = sanitize.expect_bool_or_None,
    althea_failed = sanitize.expect_bool_or_None,
    althea_fangirl = sanitize_bc.expect_Student_instance_or_None,
    althea_festival_split = sanitize.expect_bool_or_del,
    althea_just_friends = sanitize.expect_bool,
    althea_movie = sanitize.expect_bool,
    althea_nov_19_noisy = sanitize.expect_bool_or_None,
    althea_path_feb_1_seen = sanitize.expect_bool_or_None,
    althea_path_mallory_fluster_seen = sanitize.expect_bool_or_None,
    althea_path_moral_support_seen = sanitize.expect_bool_or_None,
    althea_path_tattoo_seen = sanitize.expect_int_or_None,
    althea_picnic = sanitize.expect_bool,
    althea_picnic_public = sanitize.expect_bool_or_del,
    althea_planetarium = sanitize.expect_bool,
    althea_pp = sanitize.expect_int,
    althea_romance_jan6 = sanitize.expect_bool_or_del,
    althea_success = sanitize_bc.scrub_althea_success,
    althea_tattoo_appt = sanitize.expect_bool_or_None,
    althea_tea = sanitize.expect_bool,
    althea_thais_quit = sanitize.expect_bool_or_None,
    althea_was_available = sanitize.expect_bool_or_del,
    assigned_targets = sanitize_bc.expect_rlist_of_Actor_instances,
    bigcalendar_hint = sanitize.expect_displayable_text,
    block_defeated = sanitize.expect_bool, # Boss
    boss_starting_minion_count = sanitize.expect_del_or_int,
    case = sanitize_bc.expect_Case_instance_or_del,
    case_serial = sanitize.expect_int,
    cases_in_play = sanitize_bc.expect_del_or_rlist_of_Case_instances,
    cases_last_week = sanitize_bc.scrub_cases_last_week,
    cases_per_week = sanitize.expect_del_or_int,
    cases_this_week = sanitize_bc.expect_rlist_of_Case_instances_or_etuple,
    cases_two_weeks_ago = sanitize_bc.expect_rlist_of_Case_instances_or_del_or_etuple,
    chosen_movie = sanitize.expect_displayable_text_or_del,
    cloned_locations = sanitize_bc.expect_del_or_rlist_of_Location_instances,
    closed_cases = sanitize_bc.expect_del_or_rlist_of_Case_instances,
    # Not all councillors are actually used here.
    coffee_current_date = sanitize_bc.expect_councillor_name_or_del,
    college_acceptances = sanitize_bc.scrub_college_acceptances,
    college_chosen = sanitize.expect_displayable_text_or_del,
    councillor = sanitize_bc.expect_Student_instance, # Bossbattle
    current_action = sanitize_bc.scrub_current_action,
    current_actor = sanitize.expect_None, # Bossbattle, probably needs refinement.
    current_actors = sanitize_bc.scrub_current_actors,
    current_bigcard = sanitize.expect_None,
    current_case = sanitize_bc.expect_Case_instance_or_del_or_None,
    current_item = sanitize.expect_None, # Bossbattle, probably needs refinement
    current_save = sanitize.discard,
    current_target = sanitize_bc.expect_Actor_instance_or_None,
    cutscene = sanitize.expect_bool_or_del,
    d_angle = sanitize.expect_del_or_int,
    d_anglestart = sanitize.expect_del_or_int,
    d_bounceval = sanitize.expect_del_or_int,
    d_faceval = sanitize.expect_del_or_int,
    d_x = sanitize.expect_del_or_int,
    d_y = sanitize.expect_del_or_int,
    damage = sanitize.expect_int, # Bossbattle
    dated = sanitize_bc.expect_councillor_name_or_del,
    day = sanitize.expect_del_or_int,
    delt = sanitize.expect_del_or_int,
    dice_ever_loaded = sanitize.expect_bool_or_None,
    difficulty = sanitize.expect_displayable_text_or_del,
    disable_pause_after_action = sanitize.expect_bool,
    disallow_cutscene_today = sanitize.expect_bool,
    easy = sanitize.expect_bool_or_None,
    eggs = sanitize.expect_int,
    elsa_ = sanitize_bc.expect_Student_instance_or_None,
    elsa_evades_mallory_althea_path = sanitize.expect_bool_or_None,
    fangirl_name = sanitize.expect_displayable_text_or_del,
    final_boss_victory = sanitize.expect_displayable_text_or_None,
    # Discarding forced_cases by default, because they're a debugging feature.
    fixed_mole_ = sanitize_bc.expect_Councillor_or_None,
    forced_cases = sanitize.new_empty_revertable_list,
    #g_target_roll_delay = sanitize.expect_float,
    # This is obsolete, it being missing in saves is harmless
    game_runtime = sanitize.expect_displayable_text_or_del, # Epilogue stats
    general_targets_pseudocase = sanitize_bc.expect_Case_instance_or_del,
    girl = sanitize_bc.expect_councillor_name_or_del,
    gpa = sanitize.expect_displayable_text_or_del, # Epilogue stats
    grades = sanitize_bc.scrub_grades,
    guilty1 = sanitize_bc.expect_councillor_name_or_del,
    guilty2 = sanitize_bc.expect_councillor_name_or_del,
    heart_set_on = sanitize.expect_displayable_text_or_del,
    hidden_cases = sanitize_bc.expect_del_or_rlist_of_Case_instances,
    highint = sanitize.expect_displayable_text_or_del,
    highobs = sanitize.expect_displayable_text_or_del,
    honey_stolen = sanitize.expect_bool_or_None,
    hope_trust = sanitize.expect_bool_or_del,
    in_final_battle = sanitize.expect_bool,
    inventory = sanitize_bc.scrub_inventory,
    inventory_choice = sanitize.expect_bool,
    karma = sanitize.expect_del_or_int,
    last_boss_attack = sanitize_bc.scrub_last_boss_attack, #Boss
    last_dropped_paper = sanitize_bc.expect_int3,
    last_fighting_in_the_ranks_weeknum = sanitize.expect_int,
    # The latest_* vars are used in the demo to display versions to update
    # to.
    latest_alpha = sanitize.expect_displayable_text_or_del,
    latest_beta = sanitize.expect_displayable_text_or_del,
    latest_stable = sanitize.expect_displayable_text_or_del,
    load_count = sanitize.expect_del_or_int,
    loaded_dice = sanitize.expect_bool_or_None,
    location1 = sanitize_bc.expect_del_or_Location_instance,
    location1_name = sanitize.expect_displayable_text_or_del_or_None,
    location2 = sanitize_bc.expect_del_or_Location_instance,
    location2_name = sanitize.expect_displayable_text_or_del_or_None,
    locations = sanitize_bc.scrub_locations,
    losses = sanitize.expect_int,
    maga = sanitize.expect_del_or_int,
    magb = sanitize.expect_del_or_int,
    magc = sanitize.expect_del_or_int,
    magd = sanitize.expect_del_or_int,
    mal_knows_tea = sanitize.expect_bool_or_del,
    mallory_ = sanitize_bc.expect_Councillor_or_del,
    mallory_boyfriend_revealed = sanitize.expect_bool_or_None,
    mallory_coffeehouse = sanitize.expect_bool,
    mallory_laundry_bump = sanitize.expect_displayable_text_or_None,
    mallory_movie = sanitize.expect_bool,
    mallory_party_location = sanitize.expect_displayable_text_or_None,
    mallory_path = sanitize.expect_bool_or_None,
    mallory_pp = sanitize.expect_float_or_int,
    mallory_picnic = sanitize.expect_bool,
    mallory_planetarium = sanitize.expect_bool,
    mallory_public_picnic = sanitize.expect_bool_or_del,
    mallory_state = sanitize.expect_displayable_text,
    mallory_tea = sanitize.expect_bool,
    mallory_thanksgiving_seen = sanitize.expect_bool_or_None,
    metamorphoses = sanitize_bc.scrub_metamorphoses,
    midterms_asked_about_traitor = sanitize.expect_bool,
    minion = sanitize_bc.expect_Councillor_or_del,
    minion_althea_search = sanitize.expect_bool,
    minion_mallory_bullying = sanitize.expect_bool,
    minion_rowan_box = sanitize.expect_bool,
    minion_thais_fetch = sanitize.expect_bool,
    minion_vonne_paper = sanitize.expect_bool,
    minions = sanitize_bc.scrub_minions,
    mole_ = sanitize_bc.expect_Councillor_or_del,
    mole_case = sanitize_bc.scrub_mole_case,
    mole_confessed = sanitize.expect_bool,
    mole_cooperates = sanitize.expect_bool,
    mole_reprieve = sanitize.expect_bool_or_del,
    mole_recruiter = sanitize.expect_displayable_text_or_del,
    mole_selected = sanitize_bc.scrub_mole_selected,
    molehill_mislead_success = sanitize.expect_bool,
    molehill_solved = sanitize.expect_bool,
    month = sanitize.expect_del_or_int,
    month_name = sanitize.expect_displayable_text_or_del,
    movie_date = sanitize_bc.expect_councillor_name_or_del,
    moving_cards = sanitize.new_empty_revertable_list,
    mymole = sanitize_bc.expect_councillor_name_or_del,
    name = sanitize.expect_displayable_text_or_del,
    notebook_last_drawn = sanitize.expect_del_or_int2,
    num_minions = sanitize.expect_int, # Bossbattle
    old_load_count = sanitize.expect_del_or_int,
    old_reputation = sanitize.expect_del_or_int,
    pending_college_acceptances = sanitize_bc.scrub_pending_college_acceptances,
    pending_minion_updates = sanitize.expect_del_or_rlist_of_displayable_text,
    PLAYER_ROLLPOS = sanitize.expect_int2, # Bossbattle
    plot = sanitize.expect_bool_or_del,
    plotless_duration = sanitize.expect_del_or_int,
    proll = sanitize.expect_int, # Bossbattle
    pstatv = sanitize.expect_int, # Bossbattle
    queen_relationship = sanitize.expect_displayable_text_or_del,
    queen_selected = sanitize_bc.expect_councillor_name_or_del,
    queue_althea_embarrassing_2 = sanitize.expect_bool_or_None,
    queue_althea_path_drunk = sanitize.expect_bool_or_None,
    queue_althea_path_thais_return = sanitize.expect_bool_or_int_or_None,
    queue_althea_path_vonne = sanitize.expect_bool_or_None,
    queue_althea_romance_planetarium = sanitize.expect_bool_or_None,
    queue_mallory_birthday_cookies = sanitize.expect_bool,
    queue_mallory_celebrate = sanitize.expect_bool_or_None,
    queue_mallory_clear_air = sanitize.expect_bool_or_None,
    queue_mallory_laundry_followup = sanitize.expect_bool_or_int_or_None,
    queue_mallory_museum = sanitize.expect_bool_or_None,
    queue_mallory_room_party = sanitize.expect_bool_or_None,
    queue_mallory_stomp_apology = sanitize.expect_bool_or_None,
    queue_minion_althea_search = sanitize.expect_bool,
    queue_minion_mallory_bullying = sanitize.expect_bool,
    queue_minion_thais_fetch = sanitize.expect_bool,
    queue_naked_mollery_rats = sanitize.expect_bool,
    queue_packard_acceptance_letter = sanitize.expect_bool_or_None,
    queue_thais_guilt_followup = sanitize.expect_bool_or_None,
    queue_thais_vonne_complaint = sanitize.expect_bool_or_None,
    queue_rowan_beta_new_look = sanitize.expect_bool,
    queue_rowan_coffee_date = sanitize.expect_bool_or_del,
    queue_rowan_gene_hunt = sanitize.expect_bool_or_None,
    queue_rowan_path_mallory = sanitize.expect_bool,
    queue_vonne_althea_nudge = sanitize.expect_bool_or_None,
    queue_vonne_big_hint = sanitize.expect_bool,
    queue_vonne_class_ring = sanitize.expect_bool,
    queue_vonne_date_resched = sanitize.expect_bool_or_None,
    queue_vonne_kiss_talk = sanitize.expect_bool_or_None,
    queue_vonne_mallory_paranoid = sanitize.expect_bool_or_None,
    queue_vonne_mallory_postdate = sanitize.expect_bool_or_None,
    queue_vonne_moral_support = sanitize.expect_bool_or_None,
    queue_vonne_restaurant_date = sanitize.expect_bool_or_None,
    queue_vonne_post_break = sanitize.expect_bool_or_None,
    queue_vonne_roleplaying = sanitize.expect_bool,
    queue_vonne_secret_rendezvous = sanitize.expect_bool_or_None,
    queue_vonne_yandere = sanitize.expect_bool_or_None,
    )

STORE_SCRUBBERS.update(
    redmask_detainee_neutralized = sanitize.expect_bool,
    redmask_failed = sanitize.expect_bool_or_None,
    redmask_queen_threatened = sanitize.expect_bool,
    redmask_queen_withdrawn = sanitize.expect_bool_or_None,
    red_mask_state = sanitize.expect_displayable_text_or_del,
    registration_key = sanitize.always_None,
    registration_key_widget = sanitize.always_del,
    rep_gain = sanitize.expect_del_or_int,
    reputation = sanitize.expect_int,
    result = sanitize_bc.scrub_result,
    rival_school = sanitize.expect_displayable_text_or_None,
    roleplay = sanitize.expect_displayable_text,
    rounds = sanitize.expect_int, # Bossbattle
    rowan_ = sanitize_bc.expect_Councillor_or_del,
    rowan_carn_check = sanitize.expect_bool_or_del,
    rowan_coffeehouse = sanitize.expect_bool,
    rowan_coffeehouse_elsa_perform = sanitize.expect_bool_or_del,
    rowan_coffeehouse_perform = sanitize.expect_bool_or_del,
    rowan_friend_council = sanitize.expect_bool_or_del,
    rowan_friend_old = sanitize.expect_bool_or_del,
    rowan_friend_roommate = sanitize.expect_bool_or_del,
    rowan_gene_hunt_leave_the_past_alone = sanitize.expect_bool_or_None,
    rowan_kissed = sanitize.expect_bool,
    rowan_movie = sanitize.expect_bool,
    rowan_offered_coffee = sanitize.expect_int,
    rowan_path = sanitize_bc.scrub_rowan_path,
    rowan_picnic = sanitize.expect_bool,
    rowan_planetarium = sanitize.expect_bool,
    rowan_pp = sanitize.expect_float_or_int,
    rowan_pushed_away = sanitize.expect_bool_or_None,
    rowan_restaurant_date = sanitize.expect_bool,
    rowan_slave = sanitize.expect_bool_or_None,
    rowan_sneak_date = sanitize.expect_bool_or_None,
    rowan_tea = sanitize.expect_bool,
    rowan_tea_gotsnack = sanitize.expect_bool_or_del,
    rowan_vonne_accusation = sanitize.expect_bool_or_None,
    rowan_vonne_challenge_seen = sanitize.expect_bool,
    rowan_vs_thais_seen = sanitize.expect_bool_or_None,
    rowans_revenge = sanitize_bc.expect_Case_instance_or_del_or_None,
    rowans_revenge_thats_all = sanitize.expect_bool,
    rowans_revenge_what_then = sanitize.expect_bool,
    s = sanitize.expect_displayable_text_or_del, # For plotless
    saturday_date = sanitize_bc.scrub_saturday_date,
    scapegoat_ = sanitize_bc.scrub_scapegoat_,
    seen_vonne_kiss_kiss = sanitize.expect_bool_or_None,
    seen_vonne_kiss_scared = sanitize.expect_bool_or_None,
    seen_vonne_letter_warmup = sanitize.expect_bool,
    seen_vonne_tutor_mallory = sanitize.expect_bool,
    should_bypass_assignments = sanitize.expect_bool_or_None,
    stat = sanitize_bc.scrub_last_boss_attack,
    # student_names_this_week is unimportant - it's just used to weight random
    # names towards dissimilar ones.
    student_names_this_week = sanitize.new_empty_revertable_list,
    students_in_play = sanitize_bc.scrub_students_in_play,
    supply_closet = sanitize_bc.scrub_supply_closet,
    suppress_no_new_information = sanitize.expect_bool,
    talmage_rescued = sanitize.expect_bool_or_None,
    talmage_reward = sanitize.expect_displayable_text_or_None,
    TARGET_ROLLPOS = sanitize.expect_int2,
    target_scroll_xoff = sanitize.expect_int,
    targets_in_play = sanitize_bc.expect_rlist_of_Actor_instances,
    thais_ = sanitize_bc.expect_Councillor_or_del,
    thais_althea_was_holding_you_back = sanitize.expect_bool,
    thais_coffeehouse = sanitize.expect_bool,
    thais_dressup_dontwanttodate = sanitize.expect_bool,
    thais_fashion_show = sanitize.expect_bool,
    thais_kissu = sanitize.expect_bool_or_del,
    thais_love_letters_seen = sanitize.expect_bool_or_None,
    thais_movie = sanitize.expect_bool,
    thais_path = sanitize.expect_bool_or_None,
    thais_picnic = sanitize.expect_bool,
    thais_planetarium = sanitize.expect_bool,
    thais_pp = sanitize.expect_int,
    thais_pretty_girl = sanitize.expect_bool,
    thais_rumours_solved = sanitize.expect_bool_or_None,
    thais_rumours_timeout = sanitize.expect_bool_or_None,
    thais_skate_blowoff = sanitize.expect_bool_or_None,
    thais_skating_dumb = sanitize.expect_bool_or_del,
    thais_tea = sanitize.expect_bool,
    thais_vonne_outing = sanitize.expect_bool_or_del_or_displayable_text,
    the_final_mystery = sanitize_bc.expect_Case_instance_or_del,
    training_points_ever = sanitize.expect_int,
    training_room = sanitize_bc.scrub_training_room,
    troll = sanitize.expect_int, # Bossbattle
    tstatv = sanitize.expect_int, # Boss
    unannounced_cases =  sanitize_bc.expect_del_or_rlist_of_Case_instances,
    unassigned_actors = sanitize_bc.expect_rlist_of_Actor_instances,
    unassigned_targets = sanitize_bc.expect_rlist_of_Actor_instances,
    underground_case = sanitize_bc.scrub_underground_case,
    unlock_althea_vonne_request = sanitize.expect_bool_or_None,
    updater_object = sanitize.always_None,
    updater_output_text = sanitize.expect_displayable_text_or_None,
    updatesloop_played = sanitize.expect_bool,
    unipire = sanitize.expect_displayable_text_or_del, # For unipire case
    use_directx10 = sanitize.expect_int,
    victim2 = sanitize.expect_displayable_text_or_del, # For unipire case
    vlovepoints = sanitize.expect_del_or_int,
    vonne_ = sanitize_bc.expect_Councillor_or_del,
    vonne_birthday_party_snub = sanitize.expect_bool_or_None,
    vonne_coffeehouse = sanitize.expect_bool,
    vonne_coffeehouse_elsa_perform = sanitize.expect_bool_or_del,
    vonne_movie = sanitize.expect_bool,
    vonne_path = sanitize_bc.scrub_vonne_path,
    vonne_picnic = sanitize.expect_bool,
    vonne_planetarium = sanitize.expect_bool,
    vonne_tea = sanitize.expect_bool,
    vonnesplitcake = sanitize.expect_bool_or_del,
    vtitle = sanitize.expect_displayable_text,
    vyanpoints = sanitize.expect_del_or_int,
    weeknum = sanitize.expect_int,
    weekday = sanitize.expect_del_or_int,
    weekday_name = sanitize.expect_displayable_text_or_del,
    wins = sanitize.expect_int, 
    x = sanitize.expect_int,
    year = sanitize.expect_del_or_int,
    )

STORE_SCRUBBERS.update(
    # Ren'Py internals.
    args = sanitize.expect_del_or_emptytuple,
    kwargs = sanitize.expect_del_or_emptydict,
    mouse_visible = sanitize.expect_bool,
    nvl_list = sanitize.scrub_nvl_list,
    save_name = sanitize.discard, # save_name variable doesn't need to be
                                  # preserved in saves, as the save name is
                                  # actually carried in SaveInfo()
    suppress_overlay = sanitize.expect_bool,
    _args = sanitize.expect_None,
    _game_menu_screen = sanitize.scrub__game_menu_screen,
    _kwargs = sanitize.expect_None,
    _last_raw_what = sanitize.expect_displayable_text_or_del_or_None,
    _last_say_what = sanitize.expect_displayable_text_or_None,
    _last_say_who = sanitize.scrub_last_say_who,
    _last_voice_play = sanitize.expect_del_or_None,
    _m1_developer__missing_scene = sanitize.expect_bool,
    _old_game_menu_screen = sanitize.scrub__game_menu_screen,
    _old_predict_screens = sanitize.new_empty_revertable_list,
    _predict_screens = sanitize.new_empty_revertable_list,
    _return = sanitize.expect_int_or_None,
    _side_image_attributes = sanitize.expect_None,
    # Spiky Ren'Py features
    debugger_used = sanitize.expect_bool,
    LAST_GPG_CTX = sanitize.always_del, # For traceback upload.
    updates_available = sanitize.new_empty_revertable_list,
    # Used by save namer, no need to load.
    keyboard_target_widget = sanitize.discard,
    # message is used by the traceback uploader.
    message = sanitize.expect_displayable_text_or_del,
    old_key_repeat = sanitize.expect_del_or_int2, # Save namer
    save_caption_widget = sanitize.discard, # Save namer.
    warnstr = sanitize.discard, # Nopickles warnings.
    )

OLD_DATEWARP_STORE_SCRUBBERS = dict(
    # Deleted all the definitely-DW-specific stuff.
    current_save = sanitize.discard,
    load_count = sanitize.expect_del_or_int_or_None,
    result = sanitize.expect_del_or_str, # Used by traceback uploader
    old_load_count = sanitize.expect_del_or_int,
    predict_screens = sanitize.new_empty_revertable_list,
    ret = sanitize.expect_del_or_str,
    save_caption_widget = sanitize.expect_del,
    save_name = sanitize.expect_str,
    warnstr = sanitize.discard,
    _windows_hidden = sanitize.expect_bool,
    )
OK_VARNAME_SUBSEQUENT = set('0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz')
OK_VARNAME_FIRST = set('ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz')

def scrub_varname(n):
    if n.startswith('__'):
        raise Exception("No replacing magic variables!")
    if n[0] not in OK_VARNAME_FIRST:
        raise Exception("Invalid character in variable name: "+n[0])
    for c in n[1:]:
        if c not in OK_VARNAME_SUBSEQUENT:
            raise Exception("Invalid character in variable name: "+c)
    return n

def scrub_tnv(line, scrubbers):
    tnv = line.split(' ',2)
    t = tnv[0]
    n = scrub_varname(tnv[1])
    # Singleton type/values first.
    if t=='None':
        rv = None
    elif t=='True':
        rv = True
    elif t=='False':
        rv = False
    elif t=='del':
        rv = renpy.python.deleted
    elif len(tnv)!=3:
        print "Debug 344:",line
        raise Exception("Malformed tnv line.")
    elif t=='str':
        rv = unescape_str(tnv[2])
    else:
        print "Flawed line was",line
        raise Exception("Unexpected tnv line.")
    if n in scrubbers:
        warnings.note('tnv '+n)
        ret = n, scrubbers[n](rv)
        warnings.unnote()
        return ret
    elif INSECURE:
        print "Debug: returning un-scrubbed entry for",line
        return n, rv
    print "No scrubber for",line
    raise Exception("No scrubber found.")

class SafeloadDynamicStackEntryContext(object):
    def __init__(self):
        self.stack = []
    def push(self):
        self.stack.append({})
    def handle(self, context_stack, line):
        if line=='end':
            context_stack.pop()
            return
        warnings.note('dynamicstack')
        k, v = scrub_tnv(line, STORE_SCRUBBERS)
        warnings.unnote()
        self.stack[-1][k] = v

def unescape_music_filename(fn):
    unescaped = sanitize.scrub_relative_path(unescape_str(fn))
    if not unescaped.endswith('.ogg'):
        raise Exception('Bad music filename.')
    return unescaped

class SafeloadMusicChannelContext(object):
    secondary_volume = 1.0
    def __init__(self, name):
        for c in name:
            if c not in 'abcdefghijklmnopqrstuvwxyz0123456789':
                raise Exception("Disallowed characters in music channel name.")
        try: # We try integer channel names first.
            self.name = int(name)
        except:
            self.name = name
        self.filenames = []
    def handle(self, context_stack, line):
        if line=='end':
            context_stack.pop()
        elif line.startswith('file '):
            self.filenames.append(unescape_music_filename(line[5:]))
        elif line.startswith('vol '):
            self.secondary_volume = float(line[4:])
        else:
            raise Exception("Unexpected line in music context.")

def scrub_store_identifier(s):
    if s=='store':
        return 'store'
    raise Exception("Unexpected store identifier "+s)

class SafeloadMusicContext(object):
    def __init__(self):
        self.channels = {}
    def handle(self, context_stack, line):
        if line.startswith('channel '):
            chan = SafeloadMusicChannelContext(line[8:])
            self.channels[chan.name] = chan
            context_stack.append(chan)
        elif line=='end':
            context_stack.pop()
        else:
            print "Line was",line
            raise Exception("Unexpected line in music context.")

class Tainted(object):
    def __init__(self, val, name=None):
        self.name = name
        self.val = val
    def deindirect_val(self):
        v = self.val
        if isinstance(v,DeindirectableContext):
            return v.deindirect()
        return v

def parse_tainted_tnv(line, context_stack):
    # This returns a tainted variable containing the line's value.
    # For values that create a new parsing context, the value will be the
    # parsing context.
    tnv = line.split(' ',2)
    t = tnv[0]
    n = tnv[1]
    if len(tnv)==3:
        v = tnv[2]
        if t=='int':
            return Tainted(int(v),n)
        elif t=='str':
            return Tainted(unescape_str(v),n)
        elif t=='float':
            return Tainted(float(v),n)
        elif t=='sro':
            ctx = SafeloadSanitizableRevertableObjectContext(v)
            context_stack.append(ctx)
            return Tainted(ctx, n)
        elif t=='sroref':
            # Find the memo.
            memoctx = get_stack_memo_context(context_stack)
            # And return the object from it.
            return Tainted(memoctx.OBJECT_MEMO[i64dec(v)], n)
    elif len(tnv)==2:
        if t=='False':
            return Tainted(False, n)
        elif t=='None':
            return Tainted(None, n)
        elif t=='True':
            return Tainted(True, n)
        elif t=='(':
            ctx = SafeloadTupleValueContext()
            context_stack.append(ctx)
            return Tainted(ctx, n)
        elif t=='[':
            ctx = SafeloadRevertableListValueContext()
            context_stack.append(ctx)
            return Tainted(ctx, n)
        elif t=='{':
            ctx = SafeloadRevertableDictValueContext()
            context_stack.append(ctx)
            return Tainted(ctx, n)
        elif t=='p{':
            ctx = SafeloadPythonDictValueContext()
            context_stack.append(ctx)
            return Tainted(ctx, n)
        elif t=='rs(':
            context_stack.append(SafeloadRevertableSetValueContext())
            return Tainted(context_stack[-1],n)
        elif t=='s(':
            context_stack.append(SafeloadPythonSetValueContext())
            return Tainted(context_stack[-1],n)
        elif t=='del':
            return Tainted(renpy.python.deleted, n)
    raise Exception('Unhandled tnv '+line)

class SafeloadStoreContext(object):
    def __init__(self):
        self.data = []
    def handle(self, context_stack, line):
        if line=='end':
            context_stack.pop()
            return
        self.data.append(parse_tainted_tnv(line, context_stack))

class SafeloadObjectsContext(object):
    onval = False
    def __init__(self):
        self.trios = []
    def handle(self, context_stack, line):
        if line=='end':
            context_stack.pop()
            return
        elif self.onval:
            self.trios[-1].append(unserialize_value(context_stack, line))
            self.onval = False
            return
        self.trios.append(line.split(' ',1))
        self.onval = True
        

class SafeloadRollbackLogEntryContext(object):
    scene = None
    dynamicstack = None
    music = None
    def __init__(self):
        self.stores = {}
        self.tainted_roots = []
        self.OBJECT_MEMO = {}

    def resolve_tainted_roots(self):
        self.untainted_roots = {}
        for t in self.tainted_roots:
            if t.name.startswith('store.'):
                varname = t.name[6:]
            else:
                raise Exception("Unexpected store variable name.")
            if varname in STORE_SCRUBBERS:
                warnings.note(('tr',varname))
                self.untainted_roots[t.name] = STORE_SCRUBBERS[varname](t.deindirect_val())
                warnings.unnote()
            elif INSECURE:
                warnings.append((6,"Accepted non-scrubbable variable",varname,t.deindirect_val()))
                self.untainted_roots[t.name] = t.deindirect_val()
            else:
                warnings.append((5,'No scrubber for '+varname+', discarded it.'))
    def handle(self, context_stack, line):
        if line=='end':
            self.resolve_tainted_roots()
            context_stack.pop()
            return
        elif line=='callstack':
            context_stack.append(SafeloadCallstackContext())
            self.callstack = context_stack[-1]
            return
        elif line.startswith('current '):
            self.current = unserialize_script_loc(line[8:])
            return
        elif line=='dynamic':
            if self.dynamicstack is None:
                self.dynamicstack = SafeloadDynamicStackEntryContext()
            self.dynamicstack.push()
            context_stack.append(self.dynamicstack)
            return
        elif line=='music':
            self.music = SafeloadMusicContext()
            context_stack.append(self.music)
            return
        #elif line=='objects':
        #    self.objects = SafeloadObjectsContext()
        #    context_stack.append(self.objects)
        #    return
        elif line=='scene':
            if self.scene is not None:
                raise Exception("RollbackLogEntry already had a scene.")
            context_stack.append(SafeloadSceneContext())
            self.scene = context_stack[-1]
            return
        elif line.startswith('store '):
            storename = scrub_store_identifier(line[6:])
            context_stack.append(SafeloadStoreContext())
            self.stores[storename] = context_stack[-1]
            return
        self.tainted_roots.append(parse_tainted_tnv(line, context_stack))

    OK_MUSIC_CHANNELS = set((7, 'music2','music3'))
    def music_update_data(self):
        if not self.music:
            return ()
        global last_music_update_data
        ret = {}
        for channelname, data in self.music.channels.iteritems():
            if channelname in self.OK_MUSIC_CHANNELS:
                ctx = renpy.audio.audio.MusicContext()
                ctx.last_filenames = data.filenames
                ctx.secondary_volume = data.secondary_volume
                ctx.last_changed = renpy.audio.audio.get_serial()
                ret[channelname] = ctx
            else:
                print "Debug 879:",channelname,data
                raise Exception("Invalid channel in save.")
            last_music_update_data = ret
        return ret
    def stores_update_data(self):
        ret = {}
        for storek, storev in self.stores.iteritems():
            if storek!='store':
                raise Exception("Cannot store stores other than store.")
            ret[storek] = r = dict()
            for item in storev.data:
                n = item.name
                if n in STORE_SCRUBBERS:
                    warnings.note('stores update '+n)
                    r[n] = STORE_SCRUBBERS[n](item.deindirect_val())
                    warnings.unnote()
                elif INSECURE:
                    warnings.append((6,"No scrubber for",n,item.val))
                    r[n] = item.deindirect_val()
                else:
                    warnings.append((5,'Discarded store item',n,'because no scrubbers findable.'))
        return ret

class SafeloadRollbackLogContext(object):
    def __init__(self):
        self.entries = []
    def handle(self, context_stack, line):
        if line=='end':
            context_stack.pop()
        elif line=='entry':
            context_stack.append(SafeloadRollbackLogEntryContext())
            self.entries.append(context_stack[-1])
        else:
            raise Exception("Unexpected line in RollbackLog context.")

# The loader works on a context stack.  Context zero is the outermost.

class SafeloadContextZero(object):
    scene = None
    log = None
    info = None
    def handle(self, context_stack, line):
        if line=='end':
            context_stack.pop()
        elif line=='log':
            if self.log is not None:
                raise Exception("Context zero already had a log.")
            context_stack.append(SafeloadRollbackLogContext())
            self.log = context_stack[-1]
        elif line=='saveinfo':
            context_stack.append(SaveInfo())
            self.info = context_stack[-1]
        elif line:
            print "Context zero rejecting line '"+line+"'"
            raise Exception("Inconsistent save.")
        # We ignore blank lines, to make savegame editing less annoying.

LOADABLE_SCREENS = set((
    # _missing_images is believed unnecessary, as it crashes on load unless I
    # save contexts, and I don't want to save contexts.
    'scheduler',
    'bigcalendar',
    'casework_althea',
    'casework_mallory',
    'casework_rowan',
    'casework_thais',
    'casework_vonne',
    'eggs_indicator',
    'moving_cards',
    'skip_day_prompt',
    'tutorial_assignments_done',
    'tutorial_assignments_harass',
    'tutorial_assignments_insensitive',
    'tutorial_assignments_question',
    'tutorial_assignments_static',
    'tutorial_assignments_selectable',
    'upsell', 'upsell_menu',
    'weekstart_plotless',
    ))

DISCARDABLE_ATS = set((
    'cardcentered', # Has animation params, we don't actually need them.
    ))

DISCARDABLE_SCREENS = set((
    '_autoreload', # Autoreload is a dev tool, we don't really need it in
                    # saves.
    '_missing_images', # Crashes on load unless I save contexts.  I don't want
                        # to save contexts.
    'exit_game_button', # Only used at the end of the demo, has special fixups
                        # around it anyhow.
    'notify', # The Ren'Py notification screen is transient.
    'checklist_notify', # The checklist notification screen is transient.
    ))

class SafeloadSceneContext(object):
    def __init__(ctx):
        ctx.layers = []
    def handle(ctx, context_stack, line):
        if line=='end':
            context_stack.pop()
        elif line.startswith('layer '):
            context_stack.append(SafeloadLayerContext(line[6:]))
            ctx.layers.append(context_stack[-1])
        else:
            print ctx,'got',line
            raise Exception("Inconsistent save.")
    def SceneLists(self):
        showing = renpy.display.image.ShownImageInfo()
        ret = renpy.display.core.SceneLists(None,showing)
        for layer in self.layers:
            for img in layer.images:
                if type(img) is not SafeloadScreenContext:
                    if img.ats:
                        at_list = img.ats
                    else:
                        at_list = []
                    if img.tag is None:
                        key = img.what
                    else:
                        key = img.tag
                    if img.tag.startswith('hide$'):
                        warnings.append((0,"Save contained a hiding displayable; discarded it."))
                        continue
                    #ret.add(layer.name, renpy.easy.displayable(img.what), at_list=at_list, key=key, name=img.what.split(' '))
                    what = img.what.split(' ')
                    if what[0]!=img.tag:
                        name = tuple([img.tag,]+what[1:])
                    else:
                        name = tuple(what)
                    ret.add(layer.name, renpy.easy.displayable(img.what), at_list=at_list, key=key, name=name)

                    showing.attributes[(layer.name, key)] = img.what.split(' ')
                else:
                    tag = img.__dict__.pop('tag')
                    if tag not in LOADABLE_SCREENS:
                        if tag in DISCARDABLE_SCREENS:
                            continue
                        elif INSECURE:
                            warnings.append((6,'Accepted screen of unknown loadability',tag,img.__dict__))
                        else:
                            warnings.append((5,'Discarded screen of unknown loadability',tag,img.__dict__))
                            continue
                    scope = img.__dict__.pop('scope',{})
                    img.__dict__.pop('name',None)
                    if img.__dict__:
                        warnings.append((5,'Screen: discarded __dict__',img.__dict__))
                    screen = renpy.display.screen.get_screen_variant(tag.split()[0])
                    try:
                        d = renpy.display.screen.ScreenDisplayable(screen, screen.tag, layer.name, widget_properties={}, scope=scope)
                        ret.add(layer.name, d, name=tag, key=tag)
                    except Exception,e:
                        warnings.append(5,'An exception occurred loading a screen, it has been discarded',tag,e)
                        continue
        return ret
# add args: (SceneLists, layername, exact_displayable, key=tag
#print "Debug:add(",self,layer,thing,key,zorder,behind,at_list,name,atl
# Debug:add( <renpy.display.core.SceneLists object at 0x373cf90> master <ATLTransform at 7fd7d9a60790> bg 0 [u'vn_overlay'] [<NamedTransform overlay_pos 0x370bf90>] (u'bg', u'rainroad') None


class SafeloadLayerContext(object):
    saybehavior = False
    def __init__(ctx, name):
        ctx.name = name
        ctx.images = [] # Also contains any screens.
    def handle(ctx, context_stack, line):
        if line=='end':
            context_stack.pop()
        elif line.startswith('img '):
            context_stack.append(SafeloadImageContext(unescape_str(line[4:])))
            ctx.images.append(context_stack[-1])
        elif line.startswith('screen '):
            context_stack.append(SafeloadScreenContext(line[7:]))
            ctx.images.append(context_stack[-1])
        elif line=='saybehavior':
            ctx.saybehavior = True
        else:
            raise Exception("Inconsistent save.")

def scrub_plainstr(s):
    for c in s:
        if c not in 'abcdefghijklmnopqrstuvwxyz':
            print "Debug:",s
            raise Exception("Invalid plainstr")
    return s

def scrub_accept_anything(s):
    if INSECURE:
        warnings.append((6,'scrub_accept_anything got',type(s),s))
        return s
    raise Exception("scrub_accept_anything only allowed in insecure mode.")

class SafeloadScreenContext(object):
    ats = None
    name = None
    # vestiges of the old sanitizers.  Now uses SCREEN_SCOPE_SCRUBBERS
    # instead.
    # sanitizable.
    #    char = scrub_accept_anything,
    #    side_image = scrub_accept_anything,
    #    what = scrub_what,
    #    what_id = scrub_plainstr,
    #    who = scrub_accept_anything,
    #    who_id = scrub_plainstr,
    #    window_id = scrub_plainstr,
    #    )
    def __init__(ctx, name):
        ctx.tag = name
        ctx.scope = dict()
    def handle(ctx, context_stack, line):
        if line=='end':
            context_stack.pop()
        elif line.startswith('name '):
            ctx.name = line[5:]
        elif line.startswith('None '):
            if ctx.name in SCREEN_SCOPE_SCRUBBERS:
                ctx.scope[varname] = SCREEN_SCOPE_SCRUBBERS[ctx.name][varname](None)
            else:
                warnings.append((5,"Discarded unscrubbable screen scope var",varname,None))
        elif line.startswith('str '):
            varname, val = line[4:].split(' ',1)
            if ctx.name in SCREEN_SCOPE_SCRUBBERS:
                ctx.scope[varname] = SCREEN_SCOPE_SCRUBBERS[ctx.name][varname](unescape_str(val))
            else:
                warnings.append((5,"Discarded unscrubbable screen scope var",line))
        else:
            raise Exception("Inconsistent save.")

class SafeloadImageContext(object):
    def __init__(ctx, tag):
        ctx.tag = tag
        ctx.ats = []
    def handle(ctx, context_stack, line):
        if line.startswith('at '):
            ctx.ats = []
            for atname in line[3:].split(','):
                trans = getattr(renpy.store, atname, None)
                if type(trans) is NamedTransform:
                    ctx.ats.append(trans)
                elif trans is None:
                    warnings.append((5,"Unexpected transform name",atname))
                else:
                    warnings.append((5,"Unexpected transform type",type(trans)))
        elif line=='end':
            context_stack.pop()
        elif line.startswith('what '):
            ctx.what = unescape_str(line[5:])
        else:
            raise Exception("Inconsistent save.")

def safeload(lines):
    zero = SafeloadContextZero()
    context_stack = [zero]

    lineno = 0
    for line in lines:
        lineno += 1
        if line.startswith('#'):
            continue
        try:
            context_stack[-1].handle(context_stack, line)
        except Exception,e:
            print "Stack was",context_stack
            print "Line was",lineno,line
            print "Exception was",e
            raise

    sls = renpy.exports.scene_lists()
    for layer in sls.layers:
        sls.layers[layer] = []
        sls.at_list[layer] = dict()

    newlog = renpy.python.RollbackLog()
    for ent in zero.log.entries:
        newent = renpy.python.Rollback()
        newlog.log.append(newent)
        newent.context = renpy.execution.Context(True)
        newent.context.runtime = zero.info.playtime
        newent.context.call_location_stack = list(ent.callstack.call_location_stack)
        newent.context.return_stack = list(ent.callstack.return_stack)
        newent.context.current = ent.current
        newent.context.scene_lists = ent.scene.SceneLists()
        newent.context.images = newent.context.scene_lists.shown
        newent.context.dynamic_stack = ent.dynamicstack.stack
        newent.context.music.update(ent.music_update_data())
        for k,v in ent.stores_update_data().iteritems():
            if k in newent.stores:
                raise Exception("Redefining an existing store.")
            newent.stores[k] = v
    newlog.current = newlog.log[-1]
    roots = zero.log.entries[-1].untainted_roots
    try:
        unroller.last_save_caption = zero.info.caption
    except:
        unroller.last_save_caption = ''
    try:
        unroller.save_started_in_version = zero.info.started_in_version
        unroller.save_played_in_versions = zero.info.played_in_versions
    except:
        unroller.save_started_in_version = 'unknown'
        unroller.save_played_in_versions = ['unknown',]
    unroller.debugger_used = unroller.debugger_used or zero.info.debugger_used
    unroller.crash_ignored = unroller.crash_ignored or zero.info.crash_ignored
    currentver = renpy.store.slurpfile('game_version')[1][0]
    if currentver not in unroller.save_played_in_versions:
        unroller.save_played_in_versions.append(currentver)
    if DEBUG:
        # Keep this around for poking.
        global last_zero
        last_zero = zero
        global last_newlog
        last_newlog = newlog
    global last_loaded_game_info
    last_loaded_game_info = zero.info
    newlog.unfreeze(roots, label='_after_load')
    raise Exception("Unfreeze failed.")

    renpy.exports.restart_interaction()

def asciize(d,recursion=0):
    ret = ''
    if isinstance(d,dict):
        if recursion>3:
            return 'Excess recursion.'
        for k, v in d.iteritems():
            ret += asciize(k,recursion+1)+':'+str(type(v))+'='+asciize(v,recursion+1)+';'
    elif type(d) is bool:
        return str(d)
    else:
        try:
            return str(d)
        except:
            return "<could not asciify "+str(type(d))+">"
    return ret

def escape_str(s):
    ret = '';
    if type(s) is not str and type(s) is not unicode:
        raise Exception("escape_str called with invalid type.")
    for c in s:
        if c=='\\':
            ret += '\\\\'
        elif c=='\n':
            ret += '\\n'
        else:
            oc = ord(c)
            if oc>=32 and oc<=126:
                ret += c
            elif oc<256:
                ret += "\\x%02x"%oc
            else:
                ret += "\\U%x;"%oc
    return str(ret)

def unescape_str(s):
    ret = ''
    i = 0
    while i<len(s):
        c = s[i]
        if c!='\\':
            ret += c
        else:
            i += 1
            c = s[i]
            if c=='\\':
                ret += '\\'
            elif c=='n':
                ret += '\n'
            elif c=='x':
                ret += unichr(int(s[i+1:i+3],16))
                i += 2
            elif c=='U':
                a = ''
                i += 1
                while s[i] in '0123456789abcdef':
                    a += s[i]
                    i += 1
                ret += unichr(int(a,16))
        i += 1
    return ret

"""
def stringify_at_block(at_list):
    ret = ''
    commaed = []
    for at in at_list:
        if type(at) is NamedTransform:
            commaed.append(at.name)
        elif at.context.context:
            if commaed:
                ret += 'at '+(','.join(commaed))+'\n'
                commaed = []
            ret += 'at '+at.varname+'(\n'
            for k,v in at.context.context.iteritems():
                ret += serialize_tnv(
        else:
            print "WARNING: Could not serialize",at
            ret += '#.1334'+str(at)+str(at.__dict__))+'\n'
    if commaed:
        return ret+'at'+(','.join(commaed))
    if ret:
        return ret[:-1]
    raise Exception("Cannot stringify empty at list.")
"""

def append_at_lists(serialized, disp, at_list):
    ats = []
    if type(disp) is renpy.display.motion.ATLTransform:
        if disp.varname:
            ats.append(disp.varname)
        else:
            serialized.append('#.1490 no varname')
    else:
        serialized.append('#.1484 '+str(type(disp)))
    for at in at_list:
        if type(at) is NamedTransform:
            if ats and ats[-1]!=at.name:
                ats.append(at.name)
        else:
            unroller.last_unstringifiable_at = (at,at_list)
            if ats and ats[-1]!=at.varname:
                ats.append(at.varname)
    if ats:
        serialized.append('at '+(','.join(ats)))

def stringify_at_list(at_list):
    ret = []
    for at in at_list:
        if type(at) is NamedTransform:
            ret.append(at.name)
        else:
            unroller.last_unstringifiable_at = (at,at_list)
            ret.append(at.varname)
    return ','.join(ret)

def serialize_scene_lists(sls):
    ret = ['scene']
    for layername in sls.layers:
        ret.append('layer '+layername)
        layer = sls.layers[layername]
        ats = sls.at_list[layername]
        for shown in layer:
            if FUTUREDEBUG:
                try:
                    ret.append('#'+repr(shown)+str(type(shown.displayable)))
                except:
                    ret.append('#Unrepresentable'+str(type(shown.displayable)))
            if shown.tag is not None:
                if shown.tag.startswith('hide$'):
                    # So we don't wind up sticking a spurious 'end' after
                    # a hide$ displayable - which would be bad because
                    # it would close the layer.  Also first, because a
                    # hide$ can be a ScreenDisplayable.
                    continue
                elif type(shown.displayable) is renpy.display.screen.ScreenDisplayable:
                    if ' '.join(shown.name) in DISCARDABLE_SCREENS:
                        # Don't bother saving screens that we won't be loading.
                        ret.append("#.1766 "+(' '.join(shown.name)))
                        continue
                    ret.append('screen '+shown.tag)
                    ret.append('#.1563'+str(type(shown.name))+';'+str(shown.name))
                    if shown.name != shown.tag:
                        if type(shown.name) is str:
                            ret.append('name '+shown.name)
                        elif type(shown.name) is tuple:
                            ret.append('name '+(' '.join(shown.name)))
                        else:
                            print "WARNING: Non-string ScreenDisplayable name?",shown.name
                            ret.append('name '+' '.join(shown.name))
                            ret.append('#.1573 BUG:'+str(type(shown.name))+str(shown.name))
                    for k,v in shown.displayable.scope.iteritems():
                        if type(v) is str or type(v) is unicode:
                            ret.append('str '+k+' '+escape_str(v))
                        elif v is None:
                            ret.append('None '+k)
                        else:
                            ret.append('# Skipped encoding'+k+' '+str(type(v)))
                elif not shown.tag.startswith('hide$'):
                    # Don't save hiding anims.
                    ret.append('#.1521')
                    ret.append('img '+escape_str(shown.tag))
                    if shown.orig_name is not None:
                        if type(shown.orig_name) is str or type(shown.orig_name) is unicode:
                            ret.append(str('what '+escape_str(shown.orig_name)))
                        else:
                            ret.append('what '+escape_str(' '.join(shown.orig_name)))
                    elif shown.name is not None:
                        ret.append(('what '+escape_str(' '.join(shown.name))))
                    else:
                        ret.append("#.688")
                    if shown.orig_name is not None and shown.orig_name!=shown.name:
                        ret.append('#.1238 '+(' '.join(shown.name)).encode('utf-8','replace'))
                append_at_lists(ret, shown.displayable, ats[shown.tag])
                #if shown.tag in ats and ats[shown.tag]:
                #    ret.append('at '+stringify_at_list(ats[shown.tag]))
                #    #ret.append(stringify_at_block(ats[shown.tag]))
                #else:
                #    if type(shown.displayable) is renpy.display.motion.ATLTransform:
                #        ret.append('#.1536'+shown.displayable.varname)
                #    ret.append('#No ats')
                ret.append('end')
            elif type(shown.displayable) is renpy.display.behavior.SayBehavior:
                ret.append('saybehavior')
            else:
                print "Debug: discarded scene entry with tag None",shown
        ret.append('end')
    ret.append('end')
    ret.append('#/scene')
    return ret

def serialize_black_closet_save_info():
    ret = ['saveinfo']
    try:
        if unroller.last_save_caption:
            ret.append('caption '+escape_str(unroller.last_save_caption))
    except:
        pass
    try:
        if renpy.store.queen_selected:
            ret.append('queen '+escape_str(renpy.store.queen_selected))
    except:
        pass
    ret.append('month '+str(renpy.store.month))
    ret.append('day '+str(renpy.store.day))
    try:
        ret.append('startver '+unroller.save_started_in_version)
    except:
        ret.append('#startver exception.')
    try:
        ret.append('playvers '+(';'.join(unroller.save_played_in_versions)))
    except:
        ret.append('#playvers exception.')
    ret.append('playtime '+str(int(renpy.exports.get_game_runtime())))
    if unroller.crash_ignored:
        ret.append('crash_ignored')
    if unroller.debugger_used:
        ret.append('debugger_used')
    try:
        if not renpy.store.plot:
            ret.append('plotless '+str(renpy.store.weeknum))
    except:
        ret.append('#plot exception.')
    ret.append('end')
    return ret

def safesave_serialize():
    global serialize_warnings
    if DEBUG:
        serialize_warnings = []
    ret = serialize_black_closet_save_info()+serialize_rollback_log(renpy.game.log)
    return ret

def NewestSave():
    newest = None
    newest_mtime = 0
    for fn in os.listdir(renpy.config.savedir):
        if fn.endswith('.bcs'):
            try:
                fsplit = fn[:-4].split('-',1)
                if fsplit[0]=='auto':
                    continue
                int(fsplit[0])
                int(fsplit[1])
                st = os.stat(os.path.join(renpy.config.savedir, fn))
                if st.st_mtime>newest_mtime:
                    newest = fn
                    newest_mtime = st.st_mtime
            except Exception,e:
                # Just quietly ignore saves that we can't figure out, they're
                # probably irrelevant.
                pass
    try:
        ns = newest.split('-')
        return (ns[0], int(ns[1][:-4]), newest_mtime)
    except:
        return None

def NewestAutoSave():
    newest = None
    newest_mtime = 0
    for fn in os.listdir(renpy.config.savedir):
        if fn.endswith('.bcs'):
            try:
                fsplit = fn[:-4].split('-',1)
                if fsplit[0]!='auto':
                    continue
                int(fsplit[1])
                st = os.stat(os.path.join(renpy.config.savedir, fn))
                if st.st_mtime>newest_mtime:
                    newest = fn
                    newest_mtime = st.st_mtime
            except Exception,e:
                print 1697,e,fn
    try:
        ns = newest.split('-')
        return (ns[0], int(ns[1][:-4]), newest_mtime)
    except:
        return None

OK_SAVEINFO_QUEENS = {
    'Althea':'althea',
    u'Thaïs':'thais',
    'Vonne':'vonne',
    'Mallory':'mallory',
    'Rowand':'rowan'}

class SaveInfo(object):
    caption = None
    mtime = None
    readable = False

    debugger_used = False
    crash_ignored = False

    # BC-specific
    queen = None
    month = None
    day = None
    plotless = None # For plotless; it'll be the weeknum if set.

    playtime = 0
    exists = False

    def __init__(self, filename=None):
        if filename:
            self.read_from_file(filename)
    def read_from_file(self, filename):
        warnings.note(("SaveInfo for",renpy.store.persistent._file_page,filename))
        try:
            fn = os.path.join(renpy.config.savedir,renpy.store.persistent._file_page+'-'+str(filename)+'.bcs')
            st = os.stat(fn)
            self.mtime = st.st_mtime
            f = open(fn, 'rb')
            line = True
            parse = False
            while line:
                line = f.readline()
                if line.endswith('\n'):
                    line = line[:-1]
                if line=='saveinfo':
                    parse = True
                elif not parse:
                    continue
                elif line=='end':
                    self.readable = True
                    warnings.unnote()
                    return
                else:
                    self.handle([None,],line)
            self.exists = True
        except Exception,e:
            pass
        warnings.unnote()
    def handle(self, context_stack, line):
        if line.startswith('caption '):
            # This assumes that caption is used in a context where Ren'Py
            # text tags are interpolated, but [variable] and %(variable)s
            # style tags are not.
            self.caption = unescape_str(line[8:]).replace('{','{{')
        elif line == 'crash_ignored':
            self.crash_ignored = True
        elif line == 'debugger_used':
            self.debugger_used = True
        elif line=='end':
            context_stack.pop()
        elif line.startswith('startver '):
            self.started_in_version = line[9:]
        elif line.startswith('playvers '):
            self.played_in_versions = line[9:].split(';')
        # BC-specific
        elif line.startswith('day '):
            self.day = int(line[4:])
        elif line.startswith('month '):
            self.month = int(line[6:])
        elif line.startswith('playtime '):
            self.playtime = int(line[9:])
        elif line.startswith('plotless '):
            self.plotless = int(line[9:])
        elif line.startswith('queen '):
            queen = unescape_str(line[6:])
            self.queen = sanitize_bc.expect_councillor_name(queen)
        else:
            print "Debug: Unexpected savegame info line",line
    def InfoStr(self):
        if self.caption:
            if self.timer:
                return self.caption+'\n'+self.timer
            else:
                return self.caption
        elif self.timer:
            return self.timer
        return ''
    def LoadableInDemo(self):
        if self.plotless:
            return False
        return self.month==9

autosave_serial = 0
def autosave_hammer(savedir):
    # For debugging, formerly autosave.
    global autosave_serial
    while True:
        try:
            f = open(os.path.join(savedir, 'auto', str(autosave_serial)+'.das'),'rb')
            autosave_serial += 1
            f.close()
        except:
            f = open(os.path.join(savedir, 'auto', str(autosave_serial)+'.das'),'wb')
            serialized = safesave_serialize()
            try:
                f.writelines('\n'.join(serialized))
            except:
                print "Debug: dumping serialization failure data to unroller.autosave_failure_data"
                unroller.autosave_failure_data = tuple(serialized)
                if DEBUG:
                    renpy.store.find_type_in_seq(unroller.autosave_failure_data,unicode)
                raise
            f.close()
            autosave_serial += 1
            return

def choose_autosave_slot():
    mtimes = {}
    for slot in '123456':
        try:
            st = os.stat(os.path.join(renpy.config.savedir,'auto-'+slot+'.bcs'))
        except Exception,e:
            print "Debug 1813:",e
            return slot
        mtimes[slot] = st.st_mtime
    oldest_slot = '1'
    oldest_mtime = mtimes.pop('1')
    for slot,t in mtimes.iteritems():
        if t<oldest_mtime:
            oldest_mtime = t
            oldest_slot = slot
    return oldest_slot

def autosave_playervisible():
    slot = choose_autosave_slot()
    renpy.store.nopickle_save('auto-'+slot)

def try_all_saves(savedir):
    files = os.listdir(savedir)
    warnings.note(("try_all_saves(",savedir,")"))
    for fn in files:
        warnings.note((fn,))
        try:
            if fn.endswith('.das') or fn.endswith('.bcs'):
                f = file(os.path.join(savedir,fn),'rb')
                lines = f.read().split('\n')
                f.close()
                safeload(lines)
        except renpy.game.RestartTopContext:
            pass
        except renpy.game.RestartContext:
            pass
        except Exception,e:
            warnings.append((0,"An exception occurred loading",fn,e))
        warnings.unnote()
    warnings.unnote()

def TEST():
    print "Debug: Test cleared",len(warnings)
    clear_warnings()
    try_all_saves(os.path.join(renpy.config.savedir,'auto'))
    print "...  and the warnings went to",len(warnings)
    global uwarnings
    uwarnings = []
    seen = set()
    for warning in warnings:
        try:
            if warning[:-2] not in seen:
                seen.add(warning[:-2])
                uwarnings.append(warning)
        except:
            uwarnings.append(warning)

def init_save_data():
    unroller.last_save_caption = ''
    ver = renpy.store.slurpfile('game_version')[0][1]
    print "Debug 1794:",ver
    unroller.save_played_in_versions = [ver,]
    partver = renpy.store.slurpfile('partial_update')
    origver = renpy.store.slurpfile('original_game_version')
    if partver:
        ver += ';part:'+partver[0][1]
    if origver:
        ver += ';orig:'+origver[0][1]
    unroller.save_started_in_version = ver

class SafeloadPersistentContext(object):
    preferences = None
    def __init__(self):
        self.data = []
    def handle(self, context_stack, line):
        if line=='end':
            context_stack.pop()
            return
        elif line=='_preferences':
            context_stack.append(SafeloadStoreContext())
            self.preferences = context_stack[-1]
            return
        else:
            self.data.append(parse_tainted_tnv(line, context_stack))
    
class Persistent(object):
    # Built-in values:
    install_beta_versions = None
    install_alpha_versions = None
    _preferences = None

    _ok_underscored = ('_chosen', '_file_page', '_seen_audio', '_seen_ever', '_seen_images', '_set_preferences', '_style_preferences', '_voice_mute')

    # Replacement for the Ren'Py persistent object.
    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        return None
    def __setattr__(self, name, val):
        if name.startswith('_') and name not in self._ok_underscored:
            warnings.append((5,"Persistent disallowed forbidden preference",name))
        if name in self.__dict__ and self.__dict__[name] == val:
            # Don't bother marking things changed if the change didn't
            # do anything.  XXX this needs to correctly handle member
            # changes.
            return
        self.__dict__[name] = val
    def __getstate__(self):
        print "Persistent.getstate"
        return ()
    def _update(self):
        raise Exception("_update deprecated.")
        if self._preferences is None:
            self.__dict__['_preferences'] = renpy.preferences.Preferences()
        if not self._seen_ever:
            self.__dict__['_seen_ever'] = {}
        if not self._seen_images:
            self.__dict__['_seen_images'] = {}
        if not self._chosen:
            self.__dict__['_chosen'] = {}
        if not self._seen_audio:
            self.__dict__['_seen_audio'] = {}
        if self._changed is None:
            self.__dict__['_changed'] = {}
    def __init__(self):
        self.__dict__['_preferences'] = renpy.preferences.Preferences()
        self.__dict__['_seen_ever'] = {}
        self.__dict__['_seen_images'] = {}
        self.__dict__['_chosen'] = {}
        self.__dict__['_seen_audio'] = {}
        self.__dict__['_changed'] = {}

def persistent_serialize(persistent):
    ret = ['persistent']
    ret.extend(serialize_tnv('install_alpha_versions',persistent.install_alpha_versions,rollobjmemo=()))
    ret.extend(serialize_tnv('install_beta_versions',persistent.install_beta_versions,rollobjmemo=()))
    if sanitize_bc.DISCARD_SEEN_IMAGES:
        # Specific to Black Closet, as persistent._seen_images is not
        # used in bc and will otherwise be cluttered with cards for every
        # random student.
        persistent._seen_images = {}
    for k,v in persistent.__dict__.iteritems():
        if k=='_changed': # This is just a list of the times members were
            # changed last.  It shouldn't be serialized.
            pass
        elif k=='_preferences':
            ret.extend(persistent_serialize_preferences(v))
        else:
            ret.extend(serialize_tnv(k,v,rollobjmemo=()))
    ret.append('end')
    return ret
def persistent_serialize_preferences(_preferences):
    ret = ['_preferences']
    for pref,val in _preferences.__dict__.iteritems():
        ret.extend(serialize_tnv(pref, val,rollobjmemo=()))
    ret.append('end')
    return ret

def scrub_persistent_changed(s):
    if s:
        warnings.append((5,"scrub_persistent_changed discarded",s))
    return {}

def scrub_persistent_crypto(s):
    if s is None:
        return None
    if s=='gpg' or s=='plaintext':
        return str(s)
    if s is False:
        return False
    warnings.append((5,'scrub_persistent_crypto discarded',s))
    return None

def scrub_datewarp_persistent_chosen(s):
    # Just silently discard the chosen because it's not used in Date Warp.
    # For other games, it will be a dict of (locationtuple):menuoption pairs.
    #if s:
    #    warnings.append((5,'scrub_persistent_chosen discarded',type(s),s))
    return {}

def scrub_persistent_seen_ever(s):
    if type(s) is not dict:
        warnings.append((5,'scrub_persistent_seen_ever discarded',type(s),s))
        return {}
    ret = {}
    for k,v in s.iteritems():
        if v is True:
            if type(k) is str:
                ret[k] = True
            elif type(k) is tuple and len(k)==3 and type(k[0]) is str and type(k[1]) is int and type(k[2]) is int:
                ret[k] = True
            else:
                warnings.append((5,'scrub_persistent_seen_ever discarded',k,v))
        else:
            warnings.append((5,'scrub_persistent_seen_ever discarded',k,v))
    return ret

def ok_persistent_single_seen_image(k,v):
    if v is not True:
        return False
    if type(k) is not tuple:
        return False
    for ik in k:
        if type(ik) is not str:
            return False
    return True

def scrub_persistent_seen_images(s):
    ret = {}
    for k,v in s.iteritems():
        if ok_persistent_single_seen_image(k,v):
            ret[k] = v
        else:
            warnings.append((5,'scrub_persistent_seen_images discarded',k,v))
    return ret

def scrub_persistent_style_preferences(s):
    if s or type(s) is not renpy.python.RevertableDict:
        warnings.append((5,'scrub_persistent_style_preferences discarded',type(s),s))
    return renpy.python.RevertableDict()

def scrub_persistent_voice_mute(s):
    if type(s) is renpy.python.RevertableSet and not s:
        return renpy.python.RevertableSet()
    warnings.append((5,'scrub_persistent_voice_mute discarded',s))
    return renpy.python.RevertableSet()

def scrub_cgs(s):
    ret = set()
    for k in s:
        if type(k) is str:
            ret.add(sanitize.scrub_relative_path(k))
        else:
            warnings.append((5,"Discarded invalid checklist item.", k))
    return ret

def scrub_checklist(s):
    ret = set()
    for k in s:
        if type(k) is str:
            ret.add(k)
        else:
            warnings.append((5,"Discarded invalid checklist item.", k))
    return ret

def scrub_persistent_language(s):
    if s is None:
        return None
    if type(s) is str:
        for c in s:
            if c not in alphalow:
                warnings.append((5,'persistent.language may only contain lowercase ASCII characters.'))
                return None
        return str(c)
    warnings.append((5,"Invalid value for persistent.language:",s))
    return None

PERSISTENT_SCRUBBERS = dict(
    # Spiky.
    art_mode = sanitize.expect_int_or_None,
    crypto = scrub_persistent_crypto,
    extra_controls = sanitize.expect_bool_or_None,
    hide_updater = sanitize.expect_bool,
    install_alpha_versions = sanitize.expect_bool_or_None,
    install_beta_versions = sanitize.expect_bool_or_None,
    language = scrub_persistent_language,
    use_visible_keyboard = sanitize.expect_bool,
    steam_achievements = sanitize.scrub_persistent_steam_achievements,
    steam_stats = sanitize_bc.scrub_persistent_steam_stats,

    # Black Closet specific:
    cgs = scrub_cgs,
    checklist = scrub_checklist,
    easy = sanitize.expect_bool_or_None,
    eggs = sanitize.expect_int_or_None,
    hide_odds = sanitize.expect_bool_or_None,
    loaded_dice = sanitize.expect_bool_or_None,    

    ## Datewarp-specific.
    #endings_reached = scrub_endings_reached,
    #wires_cheats = scrub_wires_cheats,
    # Old Ren'Py
    _changed = scrub_persistent_changed,
    _chosen = scrub_datewarp_persistent_chosen,
    _file_page = sanitize.scrub__file_page,
    _invoke_main_menu = sanitize.expect_bool,
    _seen_audio = sanitize.scrub_persistent_seen_audio,
    _seen_ever = scrub_persistent_seen_ever,
    _seen_images = scrub_persistent_seen_images,
    _set_preferences = sanitize.expect_bool,
    _style_preferences = scrub_persistent_style_preferences,
    _voice_mute = scrub_persistent_voice_mute,
    )

PREFERENCE_SCRUBBERS = dict(
    afm_after_click = sanitize.expect_bool,
    afm_enable = sanitize.expect_bool,
    afm_time = sanitize.expect_float_or_int,
    fullscreen = sanitize.expect_bool,
    joymap = sanitize.scrub_preference_joymap,
    joy_deadzone = sanitize.expect_float,
    language = sanitize.expect_None,
    mouse_move = sanitize.expect_bool,
    mute = sanitize.scrub_preference_mute,
    performance_test = sanitize.expect_bool,
    physical_size = sanitize.scrub_preference_physical_size,
    renderer = sanitize.scrub_preference_renderer,
    show_empty_window = sanitize.expect_bool,
    skip_after_choices = sanitize.expect_bool,
    skip_unseen = sanitize.expect_bool,
    transitions = sanitize.expect_int,
    text_cps = sanitize.expect_int,
    using_afm_enable = sanitize.expect_bool,
    voice_sustain = sanitize.expect_bool,
    volumes = sanitize.scrub_preference_volumes,
    wait_voice = sanitize.expect_bool,
    )
def scrub_persistent_preferences(prefdata):
    ret = renpy.preferences.Preferences()
    for item in prefdata.data:
        if item.name in PREFERENCE_SCRUBBERS:
            setattr(ret, item.name, PREFERENCE_SCRUBBERS[item.name](deindirect_if_needed(item.val)))
        else:
            warnings.append((5,'Discarded unhandled pref',item.name,item.val))
    return ret

def persistent_init():
    try:
        os.stat(os.path.join(renpy.config.savedir,'persistent.dat'))
    except:
        return Persistent()
    return persistent_load(os.path.join(renpy.config.savedir,'persistent.dat'))

def persistent_load(filename):
    if not filename.endswith('.dat'):
        print "Disabled persistent",filename
        return None
    f = open(filename,'rb')
    if f.readline() != 'persistent\n':
        warnings.append((5,"Could not load persistent: bad magic."))
        return Persistent()
    pctx = SafeloadPersistentContext()
    context_stack = [pctx]
    for line in f.readlines():
        if line.startswith('#'):
            continue
        if line.endswith('\n'):
            line = line[:-1]
        context_stack[-1].handle(context_stack,line)
        if not context_stack:
            break
    ret = Persistent()
    global last_pctx
    last_pctx = pctx
    for item in pctx.data:
        if item.name not in PERSISTENT_SCRUBBERS:
            # BlackCloset-specific
            if item.name.startswith('scene_') and deindirect_if_needed(item.val) is True:
                setattr(ret, item.name, True)
                continue
            warnings.append((5,"Discarded unhandled persistent item",item.name,item.val))
        else:
            val = PERSISTENT_SCRUBBERS[item.name](deindirect_if_needed(item.val))
            setattr(ret, item.name, val)
    ret.__dict__['_preferences'] = scrub_persistent_preferences(pctx.preferences)
    return ret

PERSISTENT_BACKUP_COUNT=15
def persistent_cycle():
    i = PERSISTENT_BACKUP_COUNT
    try:
        os.unlink(os.path.join(config.savedir, 'persistent.bak'+str(i)))
    except:
        pass
    while i>0:
        i -= 1
        try:
            os.rename(
                os.path.join(config.savedir, 'persistent.bak'+str(i)),
                os.path.join(config.savedir, 'persistent.bak'+str(i+1)))
        except:
            pass
    try:
        os.rename(
            os.path.join(config.savedir, 'persistent.dat'),
            os.path.join(config.savedir, 'persistent.bak0'))
    except:
        pass

def persistent_write(self, filename):
    f = open(filename+'.new','wb')
    f.write('\n'.join(persistent_serialize(self)))
    f.close()
    try:
        os.unlink(filename)
    except:
        pass
    os.rename(filename+'.new',filename)

def persistent_save():
    if not renpy.game.persistent._changed:
        print "Persistent not changed, saving anyhow?"
        # If nothing's been changed, no need to save anything.
    persistent_cycle()
    try:
        persistent_write(renpy.game.persistent,os.path.join(os.path.expanduser(renpy.config.savedir),'persistent.dat'))
    except Exception,e:
        warnings.append((4,"Error saving persistent.dat",e))
        raise

def i64enc(i):
    if i<=0:
        if not i:
            return '0'
        raise Exception('i64 encoding only supports positive numbers.')
    ret = ''
    while i:
        ret = chr((i&0x3f)+0x30)+ret
        i >>= 6
    return ret

def i64dec(s):
    a = 0
    for c in s:
        a <<= 6
        a += ord(c)-0x30
    return a

class SanitizableRevertableObject(renpy.python.RevertableObject):
    # Subclasses must use a _member_sanitizers_ map; we don't specify one
    # here because we don't have any defaults that would make sense to inherit.
    _member_serializers_ = ()
    _discardable_members_ = ()
    def Serialize(self,memoset,infix='',memoid=None, rollobjmemo=None):
        if memoid is None:
            memoid = id(self)
        t = type(self)
        if memoid in memoset:
            return ['sroref'+infix+' '+i64enc(memoid)]

        memoset.add(memoid)
        ret = ['sro'+infix+' '+t.__name__]
        ret.append('id '+i64enc(memoid))
        if memoid in rollobjmemo:
            self = self.RolledBackClone(rollobjmemo[memoid])
        for k in self._member_sanitizers_:
            if hasattr(self,k):
                v = getattr(self, k)
                if hasattr(t,k) and (v is getattr(t,k)):
                    # Skip values which are defaults.
                    continue
                elif k in self._member_serializers_:
                    ret.extend(self._member_serializers_[k](k,v,memoset,rollobjmemo=rollobjmemo))
                else:
                    ret.extend(serialize_tnv(k,v,memoset,rollobjmemo=rollobjmemo))
        if DEBUG:
            for k in self.__dict__:
                if (k not in self._discardable_members_) and (k not in self._member_sanitizers_):
                    serialize_warnings.append("Discarded field "+str(type(self))+"."+k)
                    ret.append("#skipped "+str(type(self))+"."+k)
        ret.append('end')
        return tuple(ret)
    def RolledBackClone(self, rollbackinfo):
        try:
            ret = type(self).__new__(type(self))
            ret.rollback(rollbackinfo)
        except Exception,e:
            print "Could not serialize "+str(type(self))+": "+str(e)
            raise
        return ret
    def AfterLoad(self):
        # Null, just there so children can override it.
        pass
